#include <stdio.h>
#include <string.h>
#include <glib.h>

/* Jiri Lebl for Intro to Computability class (CS5<something>)

   Turing machine emulator, the quick and dirty code, took me about 30 minutes
   to write, another 15 to debug and make printouts, and another 15 to add
   comments:). It's using pure C and glib (either from gtk1.0 or glib1.1
   should work, and probably earlier ones, glib is a utility/container
   library which is proven to be at least twice as fast as STL, much
   smaller codewise and best of all can be used very simply from C, not to
   mention it's under LGPL)

   The input file is in the format:
   
<cur. rule> <next rule> <cur. tape character> <character to write> <direction>

   those are separated by single spaces, states are strings, characters are
   single characters, and direction is L or R character. There can be comments
   in the usual unixey fashion by putting # as the first character on the line.
   The initial state of the machine is called START
   
Example:

START START b b R
START LOOP 1 b R
LOOP LOOP 1 b R
LOOP END b b R

*/

#define BLANK 'b'
#define LEFT 'L'
#define RIGHT 'R'

/*Empty GList's for tape and rules, plus a pointer to the current
  GList node on the tape*/
GList *tape = NULL;
GList *curpos = NULL;
GList *rules = NULL;

/*The start state*/
char *state = "START";

/*Structure with info about 1 rule*/
typedef struct _Rule Rule;
struct _Rule {
	char *cur;	/*string with the current state*/
	char *next;	/*string with the next state for this rule*/
	char in;	/*what is the character that triggers this rule*/
	char out;	/*what is written back to tape*/
	char direction; /*direction as a character 'L' or 'R'*/
};

/*Read input from stdin, and store characters (casted to pointers, which
  is fine because no architecture has chars larger then pointers) in a
  (GList *)tape, if no input was read, put a blank character on the tape.
  then set the current position to the first character on the tape*/
void
readinput(void)
{
	int c;

	printf("Input: ");
	fflush(stdout);
	/*read the tape from stdin*/
	while((c=fgetc(stdin))!='\n')
		tape = g_list_append(tape,GINT_TO_POINTER(c));
	/*if nothing was read, put in a blank character*/
	if(!tape)
		tape = g_list_append(NULL,GINT_TO_POINTER(BLANK));
	curpos = tape;
}

/*reads the rules from the file and makes new Rule Structures for them and
  stuffs them into a (GList *)rules*/
void
readrules(char *file)
{
	FILE *fp;
	Rule *r;
	char buf[256];
	
	/*read the turing machine from file*/
	if(!(fp = fopen(file,"r")))
		exit(1);
	/*always have one rule allocated*/
	r = g_new(Rule,1);
	while(fgets(buf,256,fp)) {
		char s1[256],s2[256];
		if(buf[0]=='#') /*signifies a comment*/
			continue;
		/*sscanf is braindead but works for simple inputs*/
		if(sscanf(buf,"%255s %255s %c %c %c",
			  s1,s2,&r->in,&r->out,&r->direction)==5) {
			r->cur = g_strdup(s1);
			r->next = g_strdup(s2);
			rules = g_list_append(rules,r);
			/*always have one rule allocated*/
			r = g_new(Rule,1);
		}
	}
	/*we free the rule which wasn't used*/
	g_free(r);
	fclose(fp);
}

/*do a simple linear search of the rule list and return NULL or the matching
  rule structure pointer, I should have used GHashTable's for this but this
  was marginally simpler to write:)*/
Rule *
findrule(char chr, char *st)
{
	GList *li;
	for(li=rules;li;li=g_list_next(li)) {
		Rule *r = li->data;
		if(r->in == chr && strcmp(r->cur,st)==0)
			return r;
	}
	return NULL;
}

/*prints the current state of the machine*/
void
printout(void)
{
	GList *li;
	printf("\nCurrent state of machine:\n  Current state: %s\n  ",state);
	for(li=tape;li;li=g_list_next(li)) 
		fputc(GPOINTER_TO_INT(li->data),stdout);
	printf("\n  ");
	for(li=tape;li;li=g_list_next(li)) {
		if(li==curpos)
			fputc('^',stdout);
		else
			fputc(' ',stdout);
	}
	puts("");
}

/*the main function*/
int
main(int argc, char *argv[])
{
	Rule *r;
	/*read input tape from stdin*/
	readinput();
	/*read file, if no file is given use turing.in*/
	readrules(argc>1?argv[1]:"turing.in");

	/*while we can find a rule that matches the current character on
	  tape and current state, we will loop, when we can't find any,
	  we're done*/
	while((r = findrule(GPOINTER_TO_INT(curpos->data),state))) {
		/*print what we're doing*/
		printout();
		printf("Going to state %s, writing '%c', and moving %s\n\n",
		       r->next,r->out,r->direction==LEFT?"Left":"Right");
		curpos->data = GINT_TO_POINTER((int)r->out);
		state = r->next;
		/*slightly dumb logic to move left or right, and to add
		  a GList node with a blank if we reached the end*/
		if(r->direction == LEFT) {
			if(!g_list_previous(curpos))
				tape = curpos = g_list_prepend(tape,
						       GINT_TO_POINTER(BLANK));
			else
				curpos = g_list_previous(curpos);
		} else /*RIGHT*/ {
			if(!g_list_next(curpos))
				curpos = g_list_append(curpos,
						GINT_TO_POINTER(BLANK));
			curpos = g_list_next(curpos);
		}
	}

	/*print the final state of the machine and quit*/
	printout();
	return 0;
}
