#include <stdio.h>
#include <string.h>
#include <glib.h>

GList *regs = NULL;
int reglen = 0;
GList *prog = NULL;
GList *curpos = NULL;
GHashTable *labels = NULL;

typedef struct _ProgLine ProgLine;
struct _ProgLine {
	char instr; /*instruction 'Z' 'S' 'T' or 'J'*/
	int a,b; /*first two are always numeric args*/
	char *label; /*if there is a label*/
};

void
put_reg(int reg, int val)
{
	GList *li;
	reg--; /*make reg 0 based*/
	while(reg>=reglen) {
		reglen++;
		regs = g_list_append(regs,NULL);
	}

	li = g_list_nth(regs,reg);
	li->data = GINT_TO_POINTER(val);
}

int
get_reg(int reg)
{
	GList *li;
	reg--; /*make reg 0 based*/
	while(reg>=reglen) {
		reglen++;
		regs = g_list_append(regs,NULL);
	}

	li = g_list_nth(regs,reg);
	return GPOINTER_TO_INT(li->data);
}

void
readinput(void)
{
	int c;
	int i = 1;

	printf("Input (terminate args with a -1): ");
	fflush(stdout);
	/*read the tape from stdin*/
	scanf("%d",&c);
	while(c>=0) {
		put_reg(i,c);
		scanf("%d",&c);
		i++;
	}
}

void
readprog(char *file)
{
	FILE *fp;
	ProgLine *pl;
	char buf[256];
	int l=1;

	labels = g_hash_table_new(g_str_hash,g_str_equal);
	
	/*read the urm prog from file*/
	if(!(fp = fopen(file,"r")))
		exit(1);
	pl = g_new(ProgLine,1);
	while(fgets(buf,256,fp)) {
		char label[256]="";
		char label2[256]="";
		char *p;

		p = strchr(buf,'#');
		if(p) *p = '\0';
		p = strchr(buf,'\n');
		if(p) *p = '\0';

		/*scanf is braindamaged wrt %s, and being too lazy to do
		  better parsing we just live with it*/
		if(sscanf(buf,"%255s S(%d)",label,&pl->a)==2) {
			pl->instr='S';
		} else if(sscanf(buf,"S(%d)",&pl->a)==1) {
			pl->instr='S';
		} else if(sscanf(buf,"%255s Z(%d)",label,&pl->a)==2) {
			pl->instr='Z';
		} else if(sscanf(buf,"Z(%d)",&pl->a)==1) {
			pl->instr='Z';
		} else if(sscanf(buf,"%255s T(%d,%d)",
				 label,&pl->a,&pl->b)==3) {
			pl->instr='T';
		} else if(sscanf(buf,"T(%d,%d)",&pl->a,&pl->b)==2) {
			pl->instr='T';
		} else if(sscanf(buf,"%255s J(%d,%d,%255s",
				 label,&pl->a,&pl->b,label2)==4) {
			pl->instr='J';
			p = strchr(label2,')');
			if(p) *p = '\0';
			pl->label = g_strdup(label2);
		} else if(sscanf(buf,"J(%d,%d,%255s",
				 &pl->a,&pl->b,label2)==3) {
			pl->instr='J';
			p = strchr(label2,')');
			if(p) *p = '\0';
			pl->label = g_strdup(label2);
		} else
			continue;
		if(label[0]) {
			p = strchr(label,':');
			if(p) *p = '\0';
			g_hash_table_insert(labels,g_strdup(label),
					    GINT_TO_POINTER(l));
		}
		prog = g_list_append(prog,pl);
		l++;
		pl = g_new(ProgLine,1);
	}
	g_free(pl);
	fclose(fp);
}

void
printout(void)
{
	GList *li;
	printf("\nCurrent state of machine (registers):\n ");
	for(li=regs;li;li=g_list_next(li)) {
		printf(" %d",GPOINTER_TO_INT(li->data));
	}
	puts("");
}

int
main(int argc, char *argv[])
{
	readinput();
	readprog(argc>1?argv[1]:"urm.in");

	curpos = prog;

	while(curpos) {
		ProgLine *pl = curpos->data;
		int l;
		printout();
		puts("");
		
		switch(pl->instr) {
		case 'Z':
			printf("Z(%d)\n",pl->a);
			put_reg(pl->a,0);
			curpos = g_list_next(curpos);
			break;
		case 'S':
			printf("S(%d)\n",pl->a);
			put_reg(pl->a,get_reg(pl->a)+1);
			curpos = g_list_next(curpos);
			break;
		case 'T':
			printf("T(%d,%d)\n",pl->a,pl->b);
			put_reg(pl->b,get_reg(pl->a));
			curpos = g_list_next(curpos);
			break;
		case 'J':
			printf("J(%d,%d,%s)\n",pl->a,pl->b,pl->label);
			l = GPOINTER_TO_INT(g_hash_table_lookup(labels,
								pl->label));
			if(get_reg(pl->a)!=get_reg(pl->b)) {
				printf("Not jumping\n");
				curpos = g_list_next(curpos);
			} else if(l==0) {
				printf("Non-existant label: exitting\n");
				curpos = NULL;
			} else {
				printf("Jumping to: %d\n",l);
				curpos = g_list_nth(prog,l-1);
			}
			break;
		}
	}

	printout();
	return 0;
}
